_PARSERS = dict()


class Parser:
    def __init__(self, name, parser_function, is_active, is_deprecated, is_native):
        self.name = name
        self.parser_function = parser_function
        self.is_active = is_active
        self.is_deprecated = is_deprecated
        self.is_native = is_native


def parser(name, is_active=True, is_deprecated=False, is_native=True):
    def wrapper(func):
        _PARSERS[name] = Parser(name, func, is_active, is_deprecated, is_native)
        return func

    return wrapper



import sys

from vripper.enum.processingpriority import ProcessingPriority


def _pop(d, k, default=None):
    if k in d:
        return d.pop(k)
    return default


class VParams:
    def __init__(self, params):
        self.dl_threadpool_size = _pop(params, "download_threadpool_size", 5)
        self.pp_pool_size = _pop(params, "processing_pool_size", 0)
        self.max_dimension = _pop(params, "max_dimension")
        self.min_dimension = _pop(params, "min_dimension", 0)
        self.priority = _pop(params, "processing_priority", ProcessingPriority.SMALLER_FILESIZE)
        self.quality = _pop(params, "compression_quality", 65)
        self.min_filecount = _pop(params, "min_filecount", 1)
        self.max_filecount = _pop(params, "max_filecount")
        self.max_filesize = _pop(params, "max_filesize", sys.maxsize)
        self.min_filesize = _pop(params, "min_filesize", 0)
        self.acceptable_filesize = _pop(params, "acceptable_filesize")
        self.min_hitrate = _pop(params, "min_hitrate", .85)
        self.download_connect_timeout = _pop(params, "download_connect_timeout", 1)
        self.download_read_timeout = _pop(params, "download_read_timeout", 10)

        if len(params) > 0:
            excessive_keys = ",".join([k for k in params])
            raise ValueError(f"Unknown parameters: {excessive_keys}")

        assert self.dl_threadpool_size > 0
        assert self.pp_pool_size >= 0

    @property
    def is_postprocessing_enabled(self):
        return (not not self.max_dimension) or (not not self.quality)



class PostProgress:
    def __init__(self, vpost, len_images_to_download, is_postprocessing_enabled, callback):
        self.image_updates = dict()
        self.vpost = vpost
        self.len_images_to_download = len_images_to_download
        self.callback = callback
        self.is_postprocessing_enabled = is_postprocessing_enabled

    def report(self, image_update=None):
        if not self.callback:
            return

        if image_update:
            self.image_updates[image_update["filename"]] = image_update

        self.callback(self._get_payload())

    def _get_payload(self):
        complete_status = "complete" if self.is_postprocessing_enabled else "downloaded"
        len_images = len(self.vpost.images)
        len_complete = len([v for v in self.image_updates.values() if v["status"] == complete_status])
        len_failed = len([v for v in self.image_updates.values() if v["status"] == "failed"])
        len_pending = self.len_images_to_download - len(self.image_updates)

        return {
            "post_id": self.vpost.id,
            "post_title": self.vpost.title,
            "images_all": len_images,
            "images_to_download": self.len_images_to_download,
            "images_pending": len_pending,
            "images_complete": len_complete,
            "images_failed": len_failed,
            "status": "complete" if self.len_images_to_download == len_complete else "in_progress"
        }



import logging
import os

import requests
from commmons import with_prefix
from gallery_dl.exception import NoExtractorError
from requests import ConnectTimeout, ReadTimeout

from vripper.download import gdlwrapper
from vripper.error import ImagePermanentlyUnavailableError

logger = logging.getLogger("vripper")


def download_image(vimage, pdir, timeout):
    image_logger = with_prefix(logger, f"{vimage.filename} {vimage.main_url}")

    try:
        target_path = os.path.join(pdir, vimage.filename)
        _save_to_disk(image_logger, target_path, timeout, vimage)
        vimage.local_path = target_path
        return {"filename": vimage.filename, "status": "downloaded"}
    except (ConnectTimeout, ReadTimeout, requests.exceptions.ConnectionError) as e:
        image_logger.error(str(e))
    except ImagePermanentlyUnavailableError:
        vimage.is_available = False
    except:
        image_logger.exception("")

    return {"filename": vimage.filename, "status": "failed"}


def _save_to_disk(image_logger, target_path, timeout, vimage):
    """
    Downloads the image and returns the resulting local path.
    """
    try:
        gdlwrapper.download_single_image(vimage.main_url, target_path)
        image_logger.debug("Downloaded with: gdlwrapper")
    except NoExtractorError:
        underlying_url = vimage.get_underlying_url(timeout)
        with open(target_path, "wb") as fp:
            r = requests.get(underlying_url, headers={"Referer": vimage.main_url}, stream=True, timeout=timeout)
            fp.writelines(r.iter_content(1024))
        image_logger.debug("Downloaded with: native")



from uuid import uuid4

from vripper.model.vpost import VPost


class VThread:
    def __init__(self, id=None, title=None, url=None, with_empty_post=False):
        self.id = id or str(uuid4())
        self.title = title
        self.url = url
        self.posts = []

        if with_empty_post:
            self.posts.append(VPost())



from vripper.parser._common import parse
from vripper.parser._def import parser


@parser("imageporter")
def p(url, timeout):
    return parse(url, timeout, "//img[@class='pic']", unavailable_phrase="No file")



from importlib import import_module as _im
from pathlib import Path as _p

from vripper.parser._def import _PARSERS

__all__ = [
    _im(f".{f.stem}", __package__)
    for f in _p(__file__).parent.glob("*.py")
    if "__" not in f.stem
]


def get_parser_object(url: str):
    for name, parser_object in _PARSERS.items():
        if name in url and parser_object.is_active:
            return parser_object
    return None



from typing import Tuple

from commmons import html_from_url, head, get_host_url
from lxml.html import HtmlElement

from vripper.model.vimage import VImage
from vripper.model.vpost import VPost
from vripper.model.vthread import VThread


def _collect_images(page, index_offset):
    article = head(page.xpath("//div[@class='article-fulltext']"))
    for i, img in enumerate(article.xpath(".//img")):
        img_url = img.attrib["data-src"]
        yield VImage(index_in_post=index_offset + i, url=img_url)


def _get_next_href(page):
    pagination = head(page.xpath("//div[@class='pagination-list']"))
    if pagination is not None:
        found_current_page = False
        atags = pagination.xpath(".//a")
        for a in atags:
            is_current = "is-current" in a.attrib["class"]
            if is_current:
                found_current_page = True
                continue
            elif found_current_page:
                return a.attrib["href"]
    return None


def buondua_init(url: str):
    first_page = html_from_url(url)
    title = head(first_page.xpath("//div[@class='article-header']/h1")).text
    thread = VThread(title=title, url=url, with_empty_post=True)
    return thread, thread.posts[0], first_page


def buondua_collect_pages(payload: Tuple[VThread, VPost, HtmlElement]) -> VThread:
    thread, post, page = payload

    post.images.extend(_collect_images(page, len(post.images)))

    next_href = _get_next_href(page)
    if next_href is not None:
        page = html_from_url(get_host_url(thread.url) + next_href)
        return buondua_collect_pages((thread, post, page,))

    return thread



from commmons import html_from_url, md5

from vripper.model.vimage import VImage
from vripper.model.vthread import VThread


def _get_images(root, index_offset):
    for i, img in enumerate(root.xpath("//img[contains(@class,'album-photo')]")):
        url = img.attrib["data-src"]
        yield VImage(index_in_post=index_offset + i, url=url)


def v2ph_init(url):
    root = html_from_url(url)

    # id may contain filesystem-unsafe characters such as the question mark
    hashed_id = md5(url.split("/album/")[-1].split(".")[0].split("?")[0])

    thread = VThread(
        id=hashed_id,
        title=root.xpath("//h1[contains(@class, 'text-center')]")[0].text,
        with_empty_post=True
    )

    return thread, thread.posts[0], root


def v2ph_populate_images(payload):
    thread, post, root = payload

    while root is not None:
        post.images.extend(_get_images(root, index_offset=len(post.images)))

        # full contents are for premium users only
        root = None

    return thread



from commmons import html_from_url

from vripper.model.vimage import VImage
from vripper.model.vthread import VThread


def thm_init(url):
    root = html_from_url(url)

    thread = VThread(
        id=root.xpath("//article[contains(@class, 'post')]")[0].attrib["id"],
        title=root.xpath("//h1[@class='entry-title']")[0].text,
        with_empty_post=True
    )

    return thread, thread.posts[0], root


def thm_populate_images(payload):
    thread, post, root = payload

    imgs = root.xpath("//div[@class='entry-content']//img[contains(@class, 'jetpack-lazy-image')]")
    for i, img in enumerate(imgs):
        url = img.attrib["src"].split("?")[0]
        image = VImage(index_in_post=i, url=url)
        post.images.append(image)

    return thread



from commmons import html_from_url, get_host_url

from vripper.model.vimage import VImage
from vripper.model.vthread import VThread


def nlegs_init(url: str):
    root = html_from_url(url)

    thread = VThread(
        id=url.split('/')[-1].split('.')[0],
        title=root.xpath('//title')[0].text,
        with_empty_post=True
    )

    return thread, thread.posts[0], root, url


def nlegs_populate_images(payload):
    thread, post, root, thread_url = payload

    for i, atag in enumerate(root.xpath("//a[@class='thumbnail']")):
        href = atag.attrib["href"]
        image_link = get_host_url(thread_url) + href
        post.images.append(VImage(index_in_post=i, url=image_link))

    assert len(post.images) > 0
    thread.posts.append(post)

    return thread



from vripper.parser._common import parse
from vripper.parser._def import parser


@parser("pixhost.to")
def p(url, timeout):
    return parse(url, timeout, "//img[@id='image']", unavailable_phrase="Picture doesn't exist")



from vripper.parser._common import parse
from vripper.parser._def import parser


@parser("imagenimage")
def p(url, timeout):
    return parse(url, timeout, "//img[contains(@class, 'img-responsive')]")



from vripper.parser._common import parse
from vripper.parser._def import parser


@parser("imagetwist")
def p(url, timeout):
    return parse(
        url, timeout, xpath_expr="//img[contains(@class, 'img-responsive')]",
        unavailable_phrase="The image you were looking for could not be found"
    )



from commmons import html_from_url

from vripper.parser._def import parser


@parser("dirtyship")
def p(url: str, timeout):
    root = html_from_url(url, timeout=timeout)
    for atag in root.xpath("//div[@class='resolutions']//a"):
        if "full" in atag.text:
            return atag.attrib["href"]

    raise RuntimeError



import requests

from vripper.parser._common import parse
from vripper.parser._def import parser


@parser("imghit", is_active=False)
def p(url, timeout):
    def pcb(tree):
        form_nodes = tree.xpath("//*[@id='form-captcha']")
        assert len(form_nodes) == 1

        input_nodes = form_nodes[0].xpath("//input")
        payload = {i.attrib["name"]: i.attrib["value"] for i in input_nodes}
        return requests.post(url, payload, timeout=timeout)

    return parse(url, timeout, "//img", post_callback=pcb)



def get_thread_callback_payload(vthread, post_updates: dict, len_posts_to_download: int):
    len_in_progress = len([k for k, v in post_updates.items() if v["status"] == "in_progress"])
    len_complete = len([k for k, v in post_updates.items() if v["status"] == "complete"])
    len_pending = len_posts_to_download - len(post_updates)

    return {
        "thread_id": vthread.id,
        "thread_title": vthread.title,
        "posts_all": len(vthread.posts),
        "posts_to_download": len_posts_to_download,
        "posts_pending": len_pending,
        "posts_in_progress": len_in_progress,
        "posts_complete": len_complete,
        "status": "complete" if len_posts_to_download == len_complete else "in_progress"
    }


class ThreadProgress:
    def __init__(self, callback, vthread, len_posts_to_download):
        self.callback = callback
        self.vthread = vthread
        self.post_updates = dict()
        self.len_posts_to_download = len_posts_to_download

    def report(self, post_update=None):
        if not self.callback:
            return

        if post_update:
            self.post_updates[post_update["post_id"]] = post_update
        payload = get_thread_callback_payload(self.vthread, self.post_updates, self.len_posts_to_download)
        self.callback(payload)



from vripper.parser import get_parser_object


class VImage:
    def __init__(self, index_in_post, url, thumb_url=None):
        self.filename = str(index_in_post).zfill(4) + ".jpg"
        self.main_url = url
        self.thumb_url = thumb_url
        self.local_path = None  # Will update after a successful download.
        self.is_available = True  # Will change to False as needed.
        self.parser_object = get_parser_object(self.main_url.lower())

    def get_underlying_url(self, timeout):
        if self.parser_object is None:
            return self.main_url
        return self.parser_object.parser_function(self.main_url, timeout)



import logging
import os

from PIL import Image, UnidentifiedImageError
from commmons import with_prefix, get_filesize_in_bytes
from resizeimage import resizeimage

from vripper.enum.processingpriority import ProcessingPriority
from vripper.model.vparams import VParams

logger = logging.getLogger("vripper")


def _get_new_size(size, max_dimension: int):
    w, h = size
    if w > h:
        height = h * max_dimension / w
        width = max_dimension
    else:
        width = w * max_dimension / h
        height = max_dimension

    # Do not upscale
    if w <= width or h <= height:
        return None

    return int(width), int(height)


def has_enough_pixels(path, min_dimension=0):
    # Is there a cheaper way to do this check?
    try:
        with open(path, 'r+b') as f:
            with Image.open(f) as image:
                return min(image.size) > min_dimension
    except (UnidentifiedImageError, FileNotFoundError):
        pass
    return False


def process_then_return_new_path(path: str, vparams: VParams):
    tmp_path = path + ".tmp.jpg"
    if os.path.exists(tmp_path):
        # Windows FS API does not allow overwrites
        os.remove(tmp_path)
    with open(path, 'r+b') as f:
        with Image.open(f) as image:
            if vparams.max_dimension:
                new_size = _get_new_size(image.size, vparams.max_dimension)
                if new_size:
                    image = resizeimage.resize_contain(image, new_size)
            image.convert("RGB").save(tmp_path, quality=vparams.quality)
    return tmp_path


def is_valid_image(path):
    return has_enough_pixels(path, 0)


def process_with_constraints(path, vparams):
    image_logger = with_prefix(logger, path.split("/")[-1])

    if not is_valid_image(path):
        image_logger.error("Cannot read the image. This is possibly a corrupted file.")
        return

    filesize = get_filesize_in_bytes(path)
    if vparams.acceptable_filesize is not None and filesize <= vparams.acceptable_filesize:
        image_logger.debug(f"The file is already small enough. filesize={filesize}")
        return

    new_path = process_then_return_new_path(path, vparams)

    should_replace = True
    if vparams.priority == ProcessingPriority.SMALLER_FILESIZE:
        old_size = get_filesize_in_bytes(path)
        new_size = get_filesize_in_bytes(new_path)
        should_replace = new_size < old_size

        image_logger.debug(f"should_replace={should_replace} old_size={old_size} new_size={new_size}")

    if should_replace:
        os.replace(new_path, path)
    else:
        os.remove(new_path)



import requests
from lxml import html

from vripper.error import ImagePermanentlyUnavailableError

_MAX_DEPTH = 5


def get_form_data(tree) -> dict:
    inputs = tree.xpath("//form//input")
    return {
        i.attrib["name"]: i.attrib["value"] for i in inputs
        if "name" in i.attrib and "value" in i.attrib
    }


def parse(url, timeout, xpath_expr, attr="src", unavailable_phrase=None, post_callback=None):
    r = requests.get(url, timeout=timeout)
    if r.status_code == 404:
        raise ImagePermanentlyUnavailableError
    r.raise_for_status()

    def default_post_callback(t):
        payload = get_form_data(t)
        return requests.post(url, payload, headers={"Referer": url}, timeout=timeout)

    for i in range(_MAX_DEPTH):
        if unavailable_phrase and unavailable_phrase in r.text:
            raise ImagePermanentlyUnavailableError

        tree = html.fromstring(r.text)
        target_nodes = tree.xpath(xpath_expr)
        if target_nodes:
            assert len(target_nodes) == 1
            return target_nodes[0].attrib[attr]

        pcb = post_callback or default_post_callback
        r = pcb(tree)
        r.raise_for_status()

    raise RuntimeError



import requests

from vripper.parser._common import parse
from vripper.parser._def import parser


@parser("imgclick")
def p(url, timeout):
    def pcb(tree):
        input_nodes = tree.xpath("//form[@id='form-captcha']/input")
        payload = {i.attrib["name"]: i.attrib["value"] for i in input_nodes}
        return requests.post(url, payload, timeout=timeout)

    return parse(
        url, timeout, xpath_expr="//img[@class='pic']", post_callback=pcb,
        unavailable_phrase="The file you were looking for could not be found"
    )



from commmons import html_from_url

from vripper.model.vimage import VImage
from vripper.model.vthread import VThread


def sbgx_init(url):
    root = html_from_url(url)

    content = root.xpath("//div[contains(@class, 'post-body') and contains(@class, 'entry-content')]")[0]

    thread = VThread(
        id=content.attrib["id"].lstrip("post-body-"),
        title=root.xpath("//h1[contains(@class, 'post-title')]")[0].text,
        with_empty_post=True
    )

    return thread, thread.posts[0], content


def sbgx_populate_images(payload):
    thread, post, content = payload

    for i, div in enumerate(content.xpath(".//div[@class='separator']")):
        imgs = div.xpath('.//img')
        if imgs:
            img_url = imgs[0].attrib["src"]
            post.images.append(VImage(index_in_post=i, url=img_url))

    thread.posts.append(post)

    return thread



from commmons import get_host_url

from vripper.parser._common import parse
from vripper.parser._def import parser


@parser("nlegs.com")
def nleg_parser(url, timeout):
    path = parse(url, timeout, "//img[@class='img-res']")
    return get_host_url(url) + path


@parser("uuleg.com")
def uuleg_parser(url, timeout):
    return nleg_parser(url, timeout)


@parser("honeyleg.com")
def honeyleg_parser(url, timeout):
    return nleg_parser(url, timeout)


@parser("legbabe.com")
def legbabe_parser(url, timeout):
    return nleg_parser(url, timeout)


@parser("leg.cx")
def legcx_parser(url, timeout):
    return nleg_parser(url, timeout)



from typing import Tuple, List

from commmons import html_from_url, head, md5
from lxml.html import HtmlElement

from vripper.model.vimage import VImage
from vripper.model.vpost import VPost
from vripper.model.vthread import VThread


def sort_by_url(atags: List[HtmlElement]):
    def sanitize(a) -> int:
        maybe_int = a.attrib["href"].strip("/").split("-")[-1]
        try:
            return int(maybe_int)
        except ValueError:
            return 0

    return sorted(atags, key=lambda a: sanitize(a))


def dirtyship_init(url: str):
    root = html_from_url(url)

    t = VThread(id=md5(url), url=url, with_empty_post=True)

    h1 = head(root.xpath("//div[@class='content-title']//h1"))
    if h1 is None:
        t.title = h1.text

    return root, t, t.posts[0]


def dirtyship_get_images(payload: Tuple[HtmlElement, VThread, VPost]):
    root, thread, post = payload

    sorted_atgs = sort_by_url(root.xpath("//div[@class='gallery_grid']/a"))
    for i, atag in enumerate(sorted_atgs):
        url = atag.attrib["href"]
        image = VImage(index_in_post=i, url=url)
        post.images.append(image)

    return thread



from functools import reduce
from typing import List

from requests import HTTPError

from vripper.error import ImagePermanentlyUnavailableError
from vripper.forum.provider.buondua import buondua_init, buondua_collect_pages
from vripper.forum.provider.dirtyship import dirtyship_get_images, dirtyship_init
from vripper.forum.provider.nlegs import nlegs_init, nlegs_populate_images
from vripper.forum.provider.superbeautygirlx import sbgx_init, sbgx_populate_images
from vripper.forum.provider.thaihotmodels import thm_init, thm_populate_images
from vripper.forum.provider.v2ph import v2ph_init, v2ph_populate_images
from vripper.forum.provider.vipergirls import vg_get_xml, vg_process
from vripper.model.vpost import GalleryAsVPost
from vripper.model.vthread import VThread

# TODO: use decorator
_FORUM_REGISTRY = [
    (("vipergirls",), (vg_get_xml, vg_process)),
    (("nlegs", "uuleg", "honeyleg", "legbabe", "leg.cx"), (nlegs_init, nlegs_populate_images)),
    (("superbeautygirlx",), (sbgx_init, sbgx_populate_images)),
    (("v2ph.com",), (v2ph_init, v2ph_populate_images)),
    (("thaihotmodels",), (thm_init, thm_populate_images)),
    (("buondua",), (buondua_init, buondua_collect_pages)),
    (("dirtyship",), (dirtyship_init, dirtyship_get_images))
]


def chain(payload, func):
    return func(payload)


def _get_function_chain_by_url(url):
    for entry in _FORUM_REGISTRY:
        pattern: List[str] = entry[0]
        funcs = entry[1]
        if any(map(lambda substr: substr in url, pattern)):
            return funcs

    return None


def get_thread(url: str) -> VThread:
    funcs = _get_function_chain_by_url(url)

    if funcs:
        try:
            thread = reduce(chain, funcs, url)
        except HTTPError as e:
            if e.response.status_code in (403, 404):
                raise ImagePermanentlyUnavailableError(str(e))
            raise
        assert isinstance(thread, VThread)
    else:
        # IF the website is not supported, then fallback to GalleryDL
        post = GalleryAsVPost(url)
        thread = VThread()
        thread.posts.append(post)

    return thread



import logging
import os
import shutil

from commmons import create_zip

from vripper.download.gdlwrapper import configure_gallerydl
from vripper.download.postdownloader import download_post
from vripper.download.threadprogress import ThreadProgress
from vripper.enum.downloadmode import DownloadMode
from vripper.enum.outputformat import OutputFormat
from vripper.forum.threadfactory import get_thread
from vripper.model.vparams import VParams

logger = logging.getLogger("vripper")


def get_posts_to_download(posts, mode) -> list:
    if mode == DownloadMode.ALL_POSTS:
        posts_to_download = posts
    elif mode == DownloadMode.ALL_POSTS_WITH_IMAGES:
        posts_to_download = [p for p in posts if len(p.images)]
    elif mode == DownloadMode.FIRST_POST_ONLY:
        posts_to_download = posts[:1]
    else:
        raise NotImplementedError

    return posts_to_download


def download_thread(
        thread_url: str,
        dest: str,
        mode: DownloadMode = DownloadMode.FIRST_POST_ONLY,
        output_format: str = None,
        callback=None,
        **kwargs
):
    if callback:
        raise NotImplementedError

    vparams = VParams(kwargs)
    vthread = get_thread(thread_url)
    dest = os.path.abspath(dest)

    configure_gallerydl(vparams, dest)

    posts_to_download = get_posts_to_download(vthread.posts, mode)

    tp = ThreadProgress(callback, vthread, len(posts_to_download))

    tdir = os.path.join(dest, str(vthread.id))
    os.makedirs(tdir, exist_ok=True)

    filecount = 0
    for post in posts_to_download:
        filecount += download_post(post, tdir=tdir, vparams=vparams, callback=tp.report)

    final_path = tdir
    if output_format == OutputFormat.ZIP:
        zip_path = create_zip(basename=tdir, source_dir=tdir)
        final_path = os.path.join(dest, zip_path)
        logger.info(f"zip created final_path={final_path}")
        shutil.rmtree(tdir)
        logger.info(f"temporary folder deleted tdir={tdir}")

    tp.report()

    return {
        "thread_title": vthread.title,
        "path": final_path,
        "filecount": filecount
    }



from vripper.parser._common import parse
from vripper.parser._def import parser


# imgflare uses google captcha
@parser("imgflare", is_deprecated=True)
def p(url, timeout):
    return parse(url, timeout, "//img[@id='source']")



import requests
from lxml import html

from vripper.parser._def import parser


@parser("imagevenue")
def p(url, timeout):
    r = requests.get(url, timeout=timeout)
    tree = html.fromstring(r.text)

    img_nodes = tree.xpath("//img")
    for img in img_nodes:
        if "alt" in img.attrib:
            return img.attrib["src"]

    raise RuntimeError



from vripper.parser._common import parse
from vripper.parser._def import parser


@parser("imagezilla")
def p(url, timeout):
    return parse(url, timeout, "//img[@id='photo']", unavailable_phrase="The requested image does not exist")



from vripper.parser._common import parse
from vripper.parser._def import parser


@parser("imgbox.com")
def p(url, timeout):
    return parse(
        url, timeout, "//img[@id='img']", unavailable_phrase="The image in question does not exist"
    )



import setuptools

from generate_supported_host_markdown import get_bullet_points

with open("README.md", "r") as fh:
    long_description = fh.read()

long_description = long_description.replace(" [here](vripper/parser).", ":\n\n" + get_bullet_points())

setuptools.setup(
    name="vripper",
    version="0.5.37",
    author="vka",
    description="A Python implementation of VRipper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(exclude=("test",)),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=["requests", "lxml", "python-resize-image", "gallery-dl", "commmons"]
)



import urllib.parse as urlparse
import xml.etree.ElementTree as ET

import requests
from commmons import breakdown, get_query_params

from vripper.model.vimage import VImage
from vripper.model.vpost import VPost
from vripper.model.vthread import VThread

_API_BASE_URL = "https://vipergirls.to/vr.php"


def _check_permissions(root):
    error = root.find("error")
    if error is not None and error.attrib["type"] == "permissions":
        raise PermissionError(error.attrib["details"])


def _populate_images(post, post_element):
    for i, img_node in enumerate(post_element.findall("image")):
        img = VImage(index_in_post=i, url=img_node.attrib["main_url"], thumb_url=img_node.attrib.get("thumb_url"))
        post.images.append(img)


def vg_get_xml(url):
    thread = VThread()
    thread.id = url.split('/')[-1].split('-')[0]

    params = {"t": thread.id}
    postid = get_query_params(url).get("p")
    if postid:
        params = {"p": postid}

    r = requests.get(_API_BASE_URL, params)
    r.raise_for_status()

    root = ET.fromstring(r.text)
    _check_permissions(root)

    t = root.find("thread")
    thread.title = t.attrib["title"]
    thread.url, query_params = breakdown(url)
    reply_index = query_params.get("r")
    if reply_index:
        thread.id += f"-r{reply_index}"
        thread.title += f" Reply {reply_index}"

    return thread, root, reply_index, postid


def vg_process(payload):
    thread, root, reply_index, postid = payload

    for i, post_element in enumerate(root.findall("post")):
        if reply_index is not None and i != int(reply_index):
            continue

        post = VPost()
        post.id = post_element.attrib["id"]
        post.url = f"{thread.url}?{urlparse.urlencode({'p': post.id})}"
        post.title = thread.title
        if postid:
            post.title += f" Post {postid}"
            thread.title = post.title

        _populate_images(post, post_element)

        thread.posts.append(post)

    return thread



import logging

logger = logging.getLogger("vripper")

if not len(logger.handlers):
    # Setup logger
    logger.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)



class VPost:
    def __init__(self):
        self.id = "0"
        self.title = None
        self.images = []
        self.url = None
        self.is_gallery = False


class GalleryAsVPost(VPost):
    def __init__(self, url):
        VPost.__init__(self)
        self.url = url
        self.is_gallery = True



class UnsupportedHostError(Exception):
    """Raised when the host is unsupported (yet)"""
    pass


class DeprecatedHostError(Exception):
    """Raised when the host is unsupported (permanently)"""
    pass


class InsufficientFileCountError(Exception):
    """Raised when the final file count in a post does not meet the required value"""
    pass


class ImagePermanentlyUnavailableError(Exception):
    """Raised when the image was permanently deleted"""
    pass


class NoNativeParserError(Exception):
    """
    Raised when the underlying URL for an image cannot be determined by the native parser.
    This exception acts as the signal to use the secondary parser (ie. 3rd party).
    """
    pass



import logging
import multiprocessing
import os
from functools import partial
from multiprocessing.pool import Pool, ThreadPool

from commmons import linear_sample, get_filesize_in_bytes, with_prefix

from vripper.download.gdlwrapper import get_images_from_gallery
from vripper.download.imagedownloader import download_image
from vripper.download.postprogress import PostProgress
from vripper.error import InsufficientFileCountError, ImagePermanentlyUnavailableError, \
    DeprecatedHostError
from vripper.model.vparams import VParams
from vripper.postprocess.imgproc import process_with_constraints, has_enough_pixels, is_valid_image

logger = logging.getLogger("vripper")


def _trim(images, max_filecount):
    if max_filecount and len(images) > max_filecount:
        images_to_download = linear_sample(images, max_filecount)
    else:
        images_to_download = images

    return images_to_download


def _download_multithread(images_to_download, pdir, report_progress, vparams):
    timeout = (vparams.download_connect_timeout, vparams.download_read_timeout,)
    pool = ThreadPool(vparams.dl_threadpool_size)
    partial_func = partial(download_image, pdir=pdir, timeout=timeout)
    rs = pool.imap_unordered(partial_func, images_to_download)

    for image_update in rs:
        report_progress(image_update)

    pool.close()
    pool.join()


def _process_with_pool(paths, vparams, report_progress):
    cc = multiprocessing.cpu_count()
    if vparams.pp_pool_size > cc:
        # Warn the user if they use processing_pool_size > cc.
        logger.warning(f"processing_pool_size={vparams.pp_pool_size} is larger than cpu_count={cc}")
    pool = Pool(vparams.pp_pool_size)
    partial_func = partial(process_with_constraints, vparams=vparams)
    rs = pool.imap_unordered(partial_func, paths)

    for image_update in rs:
        report_progress(image_update)

    pool.close()
    pool.join()


def _raise_for_initial_hitrate(images_to_download, vparams):
    if len(images_to_download) == 0:
        raise ImagePermanentlyUnavailableError("The post has zero images")
    if len(images_to_download) < vparams.min_filecount:
        raise InsufficientFileCountError(f"Need {vparams.min_filecount} images. Found {len(images_to_download)}.")
    _raise_for_deprecated_hosts(images_to_download, vparams)


def _raise_for_deprecated_hosts(images_to_download, vparams):
    deprecated_hosts = [vi.parser_object.name for vi in images_to_download if
                        vi.parser_object is not None and vi.parser_object.is_deprecated]
    hitrate = 1 - len(deprecated_hosts) / len(images_to_download)
    if hitrate < vparams.min_hitrate:
        raise DeprecatedHostError(f"{set(deprecated_hosts)=}")


def _raise_for_availability(images, vparams):
    # After the download, .is_available is updated
    len_unavailable = [vimage for vimage in images if not vimage.is_available]
    if (1 - vparams.min_hitrate) < len(len_unavailable) / len(images):
        msg = f"At least {len(len_unavailable)} out of {len(images)} images are unavailable"
        raise ImagePermanentlyUnavailableError(msg)


def _postprocess(pathmap, vparams, pprog):
    # Resize the images if applicable
    if vparams.is_postprocessing_enabled:
        if vparams.pp_pool_size > 0:
            _process_with_pool(pathmap.values(), vparams, pprog.report)
        else:
            for path in pathmap.values():
                process_with_constraints(path, vparams=vparams)


def _delete_invalid_files(pathmap, vparams):
    # Delete invalid/large files
    for filename, path in pathmap.items():
        image_logger = with_prefix(logger, filename)

        if not os.path.exists(path):
            image_logger.warning("File does not exist")
            continue

        should_delete = False
        filesize = get_filesize_in_bytes(path)

        if filesize > vparams.max_filesize or filesize < vparams.min_filesize:
            image_logger.debug("Filesize too big or too small")
            should_delete = True
        elif not is_valid_image(path):
            image_logger.warning("Invalid image")
            should_delete = True
        elif not has_enough_pixels(path, vparams.min_dimension):
            image_logger.warning("The resolution is too low")
            should_delete = True

        if should_delete:
            os.remove(path)


def _evaluate_final_hitrate(images, pathmap, vparams):
    final_filecount = len([path for path in pathmap.values() if os.path.exists(path)])
    final_hitrate = final_filecount / len(images)
    if final_hitrate < vparams.min_hitrate:
        msg = f"Hitrate requirements not met; final_hitrate={final_hitrate}"
        raise InsufficientFileCountError(msg)
    return final_filecount


def download_post(vpost, tdir, vparams: VParams, callback):
    pdir = os.path.join(tdir, str(vpost.id))
    os.makedirs(pdir, exist_ok=True)

    if vpost.is_gallery:
        vpost.images = get_images_from_gallery(vpost.url)

    images = _trim(vpost.images, vparams.max_filecount)
    _raise_for_initial_hitrate(images, vparams)

    pprog = PostProgress(vpost, len(images), vparams.is_postprocessing_enabled, callback)
    pprog.report()

    # Download the images
    _download_multithread(images, pdir, pprog.report, vparams)

    # Post-download activities
    pathmap = {vimage.filename: vimage.local_path for vimage in images if vimage.local_path}
    _raise_for_availability(images, vparams)
    _postprocess(pathmap, vparams, pprog)
    _delete_invalid_files(pathmap, vparams)
    return _evaluate_final_hitrate(images, pathmap, vparams)



import logging
import os

import gallery_dl
from gallery_dl.exception import AuthorizationError, NotFoundError
from gallery_dl.extractor.message import Message
from gallery_dl.job import DownloadJob, UrlJob

from vripper.error import ImagePermanentlyUnavailableError
from vripper.model.vimage import VImage
from vripper.model.vparams import VParams


def configure_gallerydl(vparams: VParams, dest: str):
    """
    Configures gdl. Only needs to run once in each process. The configs are shared across threads.
    :param vparams:
    :param dest:
    :return:
    """

    # Set download timeout
    timeout = (vparams.download_connect_timeout, vparams.download_read_timeout,)
    gallery_dl.config.set(("downloader", "http",), "timeout", timeout)

    # Set download path
    gallery_dl.config.set((), "extractor", {"base-directory": dest})

    # Mute internal logs
    gallery_dl.config.set(("output",), "mode", "null")
    logging.getLogger("gallery-dl").setLevel(logging.CRITICAL)
    logging.getLogger("download").setLevel(logging.CRITICAL)


def _get_url_from_msg(msg):
    if msg[0] == Message.Url:
        _, url, kwdict = msg
        return url
    return None


def _get_image_urls(gallery_url):
    urlj = UrlJob(gallery_url)
    urlj.run()

    image_urls = list()
    for msg in urlj.extractor:
        image_url = _get_url_from_msg(msg)
        if image_url:
            image_urls.append(image_url)
    return image_urls


def get_images_from_gallery(gallery_url):
    try:
        image_urls = _get_image_urls(gallery_url)
    except (AuthorizationError, NotFoundError) as e:
        raise ImagePermanentlyUnavailableError(str(e))
    return [
        VImage(index_in_post=i, url=url)
        for i, url in enumerate(image_urls)
    ]


def download_single_image(url, target_path):
    dlj = DownloadJob(url)

    base_dir = gallery_dl.config.get((), "extractor")["base-directory"]
    target_path_rel = target_path.lstrip(base_dir)

    dlj.extractor.directory_fmt = tuple(os.path.dirname(target_path_rel).split(os.path.sep))
    dlj.extractor.filename_fmt = "{filename}.{extension}"
    dlj.run()

    path = os.path.join(os.path.dirname(target_path), dlj.pathfmt.filename)
    if not os.path.exists(path):
        raise FileNotFoundError(path)

    os.replace(path, target_path)
